@@ -13,9 +13,16 @@ import click
 import sqliteparser
 from tabulate import tabulate
 
-from . import Database, migrations
+from . import Database, Schema, migrations
 
 # Help strings used in multiple places.
+HELP_AUTO_TIMESTAMP_CREATE = (
+    "Automatically populate `created_at` and `last_updated_at` columns with current "
+    + "time."
+)
+HELP_AUTO_TIMESTAMP_UPDATE = (
+    "Automatically populate `last_updated_at` column with current time."
+)
 HELP_COLUMNS = "Only display these columns in the results."
 HELP_DESC = (
     "When combined with --order-by, order the results in descending rather than "
@@ -24,6 +31,7 @@ HELP_DESC = (
 HELP_HIDE = "Hide these columns in the results."
 HELP_LIMIT = "Limit the number of rows returned from the database."
 HELP_NO_CONFIRM = "Do not prompt for confirmation."
+HELP_NO_RENAME = "Don't rename columns, only add or remove them."
 HELP_OFFSET = "Offset a query with --limit."
 HELP_ORDER_BY = "Order the results by one or more columns."
 HELP_PAGE = (
@@ -86,6 +94,63 @@ def main_count(db_path, table, *, where, distinct):
         print(db.count(table, where=where, distinct=distinct))
 
 
+@cli.command(name="create")
+@click.argument("db_path")
+@click.argument("table")
+@click.argument("payload", nargs=-1)
+@click.option(
+    "--auto-timestamp/--no-auto-timestamp",
+    default=True,
+    help=HELP_AUTO_TIMESTAMP_CREATE,
+)
+def main_create(db_path, table, payload, *, auto_timestamp=True):
+    """
+    Create a new row non-interactively.
+
+    PAYLOAD should be a list of space-separated key-value pairs, e.g.
+
+        isqlite create --db db.sqlite3 my_table a=1 b=2
+    """
+    _create(db_path, table, payload, auto_timestamp=auto_timestamp)
+
+
+@cli.command(name="insert")
+@click.argument("db_path")
+@click.argument("table")
+@click.argument("payload", nargs=-1)
+@click.option(
+    "--auto-timestamp/--no-auto-timestamp",
+    default=True,
+    help=HELP_AUTO_TIMESTAMP_CREATE,
+)
+def main_insert(db_path, table, payload, *, auto_timestamp=True):
+    """
+    Alias for 'create'.
+    """
+    _create(db_path, table, payload, auto_timestamp=auto_timestamp)
+
+
+def _create(db_path, table, payload, *, auto_timestamp):
+    if not payload:
+        report_error_and_exit("payload must not be empty")
+
+    payload_as_map = {}
+    for key_value in payload:
+        key, value = key_value.split("=", maxsplit=1)
+        payload_as_map[key] = value
+
+    if auto_timestamp:
+        auto_timestamp_columns = ["created_at", "last_updated_at"]
+    else:
+        auto_timestamp_columns = []
+
+    with Database(db_path) as db:
+        pk = db.insert(
+            table, payload_as_map, auto_timestamp_columns=auto_timestamp_columns
+        )
+        print(f"Row {pk} created.")
+
+
 @cli.command(name="create-table")
 @click.argument("db_path")
 @click.argument("table")
@@ -169,6 +234,12 @@ def main_delete(db_path, table, pk, *, where, no_confirm):
 @click.argument("db_path")
 @click.argument("schema_path")
 @click.option("--table", default=None)
+@click.option(
+    "--no-rename",
+    is_flag=True,
+    default=False,
+    help=HELP_NO_RENAME,
+)
 def main_diff(db_path, schema_path, table):
     """
     Diff the database against the Python schema.
@@ -178,8 +249,8 @@ def main_diff(db_path, schema_path, table):
         _diff(db, schema, table)
 
 
-def _diff(db, schema, table):
-    diff = db.diff(schema, table=table)
+def _diff(db, schema, table, *, detect_renaming=True):
+    diff = db.diff(schema, table=table, detect_renaming=detect_renaming)
     if not diff:
         print("Nothing to migrate: database matches schema.")
         return None
@@ -297,211 +368,142 @@ def main_get(db_path, table, pk, *, plain_foreign_keys):
     """
     with Database(db_path, readonly=True) as db:
         row = db.get_by_pk(table, pk, get_related=not plain_foreign_keys)
+        if row is None:
+            report_error_and_exit(f"row {pk} not found in table {table!r}.")
+
         if not plain_foreign_keys:
             for key, value in row.items():
                 if isinstance(value, collections.OrderedDict):
                     row[key] = get_column_as_string(value)
 
-        if row is None:
-            print(f"Row {pk} not found in table {table!r}.")
-        else:
-            prettyprint_row(row)
+        prettyprint_row(row)
 
 
-@cli.command(name="insert")
+@cli.command(name="icreate")
 @click.argument("db_path")
 @click.argument("table")
-@click.argument("payload", nargs=-1)
 @click.option(
     "--auto-timestamp/--no-auto-timestamp",
     default=True,
-    help=(
-        "Automatically populate `created_at` and `last_updated_at` columns with "
-        + "current time."
-    ),
+    help=HELP_AUTO_TIMESTAMP_CREATE,
 )
-def main_insert(db_path, table, payload, *, auto_timestamp=True):
+def main_icreate(db_path, table, *, auto_timestamp=True):
     """
-    Create a new row non-interactively.
-
-    PAYLOAD should be a list of space-separated key-value pairs, e.g.
-
-        isqlite insert --db db.sqlite3 my_table a=1 b=2
+    Interactively insert a new row into the table.
     """
-    if not payload:
-        report_error_and_exit("payload must not be empty")
-
-    payload_as_map = {}
-    for key_value in payload:
-        key, value = key_value.split("=")
-        payload_as_map[key] = value
-
     if auto_timestamp:
         auto_timestamp_columns = ["created_at", "last_updated_at"]
     else:
         auto_timestamp_columns = []
 
     with Database(db_path) as db:
-        pk = db.insert(
-            table, payload_as_map, auto_timestamp_columns=auto_timestamp_columns
-        )
+        payload = _prompt_for_table(db, table, None, auto_timestamp=auto_timestamp)
+        pk = db.insert(table, payload, auto_timestamp_columns=auto_timestamp_columns)
         print(f"Row {pk} created.")
 
 
-@cli.command(name="migrate")
-@click.argument("db_path")
-@click.argument("schema_path")
-@click.option("--table", default=None)
-@click.option(
-    "--no-confirm",
-    is_flag=True,
-    default=False,
-    help="Perform the migration without requiring human confirmation.",
-)
-@click.option(
-    "--no-backup",
-    is_flag=True,
-    default=False,
-    help="Don't create a backup of the database.",
-)
-@click.option(
-    "--debug",
-    is_flag=True,
-    default=False,
-    help="Run the database in debug mode.",
-)
-def main_migrate(db_path, schema_path, table, *, no_confirm, no_backup, debug):
-    """
-    Migrate the database to match the Python schema.
-    """
-    schema = get_schema_from_path(schema_path)
-    with Database(db_path, transaction=False, debug=debug) as db:
-        diff, data_dropped = _diff(db, schema, table)
-        if diff is None:
-            return
-
-        if not no_confirm:
-            prompt = "Do you wish to perform the migration?"
-            if data_dropped:
-                prompt += red(" Some data may be lost.")
-
-            print()
-            if not click.confirm(prompt):
-                print()
-                print("Migration aborted.")
-                sys.exit(2)
-
-        if not no_backup:
-            _, backup_name = tempfile.mkstemp(
-                prefix="isqlite-backup-", suffix=".sqlite3"
-            )
-            shutil.copy2(db_path, backup_name)
-
-        try:
-            db.apply_diff(diff)
-        except Exception:
-            traceback.print_exc()
-
-            print()
-            print("Migration rolled back due to Python exception.")
-            sys.exit(2)
-
-        if no_backup:
-            print("Migration completed successfully.")
-        else:
-            print(
-                "Migration completed successfully. "
-                + f"Backup of database before migration saved at {backup_name}."
-            )
-
-
-@cli.command(name="rename-column")
+@cli.command(name="iupdate")
 @click.argument("db_path")
 @click.argument("table")
-@click.argument("old_name")
-@click.argument("new_name")
-def main_rename_column(db_path, table, old_name, new_name):
+@click.argument("pk", type=int)
+@click.option(
+    "--auto-timestamp/--no-auto-timestamp",
+    default=True,
+    help=HELP_AUTO_TIMESTAMP_UPDATE,
+)
+def main_iupdate(db_path, table, pk, *, auto_timestamp=True):
     """
-    Rename a column.
+    Interactively update a row in the table.
     """
+    if auto_timestamp:
+        auto_timestamp_columns = ["last_updated_at"]
+    else:
+        auto_timestamp_columns = []
+
     with Database(db_path) as db:
-        db.rename_column(table, old_name, new_name)
-        print(f"Column {old_name!r} renamed to {new_name!r} in table {table!r}.")
+        row = db.get_by_pk(table, pk)
+        if row is None:
+            report_error_and_exit(f"row {pk} not found in table {table!r}.")
+
+        print(
+            "Enter a blank line to leave a row unchanged. Enter NULL to set a row blank."
+        )
+        print()
+
+        payload = _prompt_for_table(db, table, row, auto_timestamp=auto_timestamp)
+
+        db.update_by_pk(
+            table, pk, payload, auto_timestamp_columns=auto_timestamp_columns
+        )
+        print(f"Row {row['id']} updated.")
 
 
-@cli.command(name="rename-table")
-@click.argument("db_path")
-@click.argument("table")
-@click.argument("new_name")
-def main_rename_table(db_path, table, new_name):
-    """
-    Rename a table.
-    """
-    with Database(db_path) as db:
-        db.rename_table(table, new_name)
-        print(f"Table {table!r} renamed to {new_name!r}.")
+def _prompt_for_table(db, table, row, *, auto_timestamp):
+    updating = bool(row is not None)
 
+    payload = {}
+    for column in db.schema[table].columns:
+        # Skip primary key columns as they will be filled in automatically.
+        if any(
+            isinstance(constraint, sqliteparser.ast.PrimaryKeyConstraint)
+            for constraint in column.definition.constraints
+        ):
+            continue
 
-@cli.command(name="reorder-columns")
-@click.argument("db_path")
-@click.argument("table")
-@click.argument("columns", nargs=-1)
-def main_reorder_columns(db_path, table, columns):
-    """
-    Change the order of columns in a table.
-    """
-    with Database(db_path) as db:
-        db.reorder_columns(table, columns)
-        print(f"Columns of table {table!r} reordered.")
+        # If auto timestamp support is enabled, skip `created_at` and
+        # `last_updated_at` columns.
+        if auto_timestamp and column.name in ("created_at", "last_updated_at"):
+            continue
 
+        required = any(
+            isinstance(constraint, sqliteparser.ast.NotNullConstraint)
+            for constraint in column.definition.constraints
+        )
 
-@cli.command(name="schema")
-@click.argument("db_path")
-@click.argument("table", required=False, default=None)
-@click.option("--as-python", is_flag=True, default=False)
-def main_schema(db_path, table, *, as_python):
-    """
-    List the names of the tables in the database.
-    """
-    with Database(db_path, readonly=True) as db:
-        if table is not None:
-            row = db.get(
-                "sqlite_master",
-                where="type = 'table' AND name = :table",
-                values={"table": table},
-            )
-            sql = row["sql"]
+        prompt_builder = [column.name, " (", column.definition.type]
+        if required:
+            prompt_builder.append(", required")
 
-            if as_python:
-                raise NotImplementedError
+        if not updating and column.definition.default:
+            prompt_builder.append(", default ")
+            if isinstance(column.definition.default, sqliteparser.ast.String):
+                prompt_builder.append(repr(column.definition.default.value))
+            elif isinstance(column.definition.default, sqliteparser.ast.Integer):
+                prompt_builder.append(repr(column.definition.default.value))
             else:
-                try:
-                    # Use sqliteparser to parse and pretty-print the table schema.
-                    create_table_statement = sqliteparser.parse(sql)[0]
-                    print(create_table_statement)
-                except sqliteparser.SQLiteParserError:
-                    # If sqliteparser can't parse the table schema, just directly print
-                    # what SQLite returned to us.
-                    print(sql)
-        else:
-            if as_python:
-                raise NotImplementedError
+                prompt_builder.append(repr(column.definition.default.value))
+
+        prompt_builder.append(")")
+        if row is not None:
+            prompt_builder.append(" - currently ")
+            prompt_builder.append(repr(row[column.name]))
+        prompt_builder.append("? ")
+
+        value = input("".join(prompt_builder)).strip()
+
+        if not value:
+            if updating:
+                # If updating an existing row, skip empty values.
+                continue
             else:
-                rows = db.select(
-                    "sqlite_master",
-                    where="type = 'table' AND name NOT LIKE 'sqlite_%'",
-                    order_by="name",
-                )
+                # If creating a new row, set the column to NULL if it is not a TEXT
+                # column or leave it as the empty string otherwise.
+                if column.definition.type != "TEXT":
+                    value = None
 
-            if rows:
-                print("\n".join(row["name"] for row in rows))
+        if value == "NULL":
+            value = None if column.definition.type != "TEXT" else ""
+
+        payload[column.name] = value
+
+    return payload
 
 
-@cli.command(name="search")
+@cli.command(name="list")
 @click.argument("db_path")
 @click.argument("table")
-@click.argument("query")
 @click.option("-w", "--where", default="")
+@click.option("-s", "--search")
 @click.option("--columns", multiple=True, default=[], help=HELP_COLUMNS)
 @click.option("--hide", multiple=True, default=[], help=HELP_HIDE)
 @click.option("-p", "--page", default=1, help=HELP_PAGE)
@@ -512,12 +514,12 @@ def main_schema(db_path, table, *, as_python):
 @click.option(
     "--plain-foreign-keys", is_flag=True, default=False, help=HELP_PLAIN_FOREIGN_KEYS
 )
-def main_search(
+def main_list(
     db_path,
     table,
-    query,
     *,
     where,
+    search,
     columns,
     hide,
     page,
@@ -528,13 +530,13 @@ def main_search(
     plain_foreign_keys,
 ):
     """
-    Shorthand for `select <table> -s <query>`
+    List the rows in the table, optionally filtered by a SQL clause.
     """
-    base_select(
+    _select(
         db_path,
         table,
         where=where,
-        search=query,
+        search=search,
         columns=columns,
         hide=hide,
         page=page,
@@ -577,9 +579,9 @@ def main_select(
     plain_foreign_keys,
 ):
     """
-    List the rows in the table, optionally filtered by a SQL clause.
+    Alias for 'list'.
     """
-    base_select(
+    _select(
         db_path,
         table,
         where=where,
@@ -595,7 +597,7 @@ def main_select(
     )
 
 
-def base_select(
+def _select(
     db_path,
     table,
     *,
@@ -611,7 +613,7 @@ def base_select(
     plain_foreign_keys,
 ):
     """
-    Base implementation shared by `main_search` and `main_select`.
+    Base implementation shared by `main_list`, `main_search`, and `main_select`.
     """
     with Database(db_path, readonly=True) as db:
         try:
@@ -679,6 +681,217 @@ def base_select(
             prettyprint_rows(rows, columns=columns, hide=hide, page=page)
 
 
+@cli.command(name="migrate")
+@click.argument("db_path")
+@click.argument("schema_path")
+@click.option("--table", default=None)
+@click.option(
+    "--no-confirm",
+    is_flag=True,
+    default=False,
+    help="Perform the migration without requiring human confirmation.",
+)
+@click.option(
+    "--no-backup",
+    is_flag=True,
+    default=False,
+    help="Don't create a backup of the database.",
+)
+@click.option(
+    "--no-rename",
+    is_flag=True,
+    default=False,
+    help=HELP_NO_RENAME,
+)
+@click.option(
+    "--debug",
+    is_flag=True,
+    default=False,
+    help="Run the database in debug mode.",
+)
+def main_migrate(
+    db_path, schema_path, table, *, no_confirm, no_backup, no_rename, debug
+):
+    """
+    Migrate the database to match the Python schema.
+    """
+    schema = get_schema_from_path(schema_path)
+    with Database(db_path, transaction=False, debug=debug) as db:
+        diff, data_dropped = _diff(db, schema, table, detect_renaming=not no_rename)
+        if diff is None:
+            return
+
+        if not no_confirm:
+            prompt = "Do you wish to perform the migration?"
+            if data_dropped:
+                prompt += red(" Some data may be lost.")
+
+            print()
+            if not click.confirm(prompt):
+                print()
+                print("Migration aborted.")
+                sys.exit(2)
+
+        if not no_backup:
+            _, backup_name = tempfile.mkstemp(
+                prefix="isqlite-backup-", suffix=".sqlite3"
+            )
+            shutil.copy2(db_path, backup_name)
+
+        try:
+            db.apply_diff(diff)
+        except Exception:
+            traceback.print_exc()
+
+            print()
+            print("Migration rolled back due to Python exception.")
+            sys.exit(2)
+
+        if no_backup:
+            print()
+            print("Migration completed successfully.")
+        else:
+            print()
+            print(
+                "Migration completed successfully. "
+                + f"Backup of database before migration saved at {backup_name}."
+            )
+
+
+@cli.command(name="rename-column")
+@click.argument("db_path")
+@click.argument("table")
+@click.argument("old_name")
+@click.argument("new_name")
+def main_rename_column(db_path, table, old_name, new_name):
+    """
+    Rename a column.
+    """
+    with Database(db_path) as db:
+        db.rename_column(table, old_name, new_name)
+        print(f"Column {old_name!r} renamed to {new_name!r} in table {table!r}.")
+
+
+@cli.command(name="rename-table")
+@click.argument("db_path")
+@click.argument("table")
+@click.argument("new_name")
+def main_rename_table(db_path, table, new_name):
+    """
+    Rename a table.
+    """
+    with Database(db_path) as db:
+        db.rename_table(table, new_name)
+        print(f"Table {table!r} renamed to {new_name!r}.")
+
+
+@cli.command(name="reorder-columns")
+@click.argument("db_path")
+@click.argument("table")
+@click.argument("columns", nargs=-1)
+def main_reorder_columns(db_path, table, columns):
+    """
+    Change the order of columns in a table.
+    """
+    with Database(db_path) as db:
+        db.reorder_columns(table, columns)
+        print(f"Columns of table {table!r} reordered.")
+
+
+@cli.command(name="schema")
+@click.argument("db_path")
+@click.argument("table", required=False, default=None)
+@click.option("--as-python", is_flag=True, default=False)
+def main_schema(db_path, table, *, as_python):
+    """
+    List the names of the tables in the database.
+    """
+    with Database(db_path, readonly=True) as db:
+        if table is not None:
+            row = db.get(
+                "sqlite_master",
+                where="type = 'table' AND name = :table",
+                values={"table": table},
+            )
+            if row is None:
+                report_error_and_exit(f"table {table!r} not found.")
+
+            sql = row["sql"]
+
+            if as_python:
+                raise NotImplementedError
+            else:
+                try:
+                    # Use sqliteparser to parse and pretty-print the table schema.
+                    create_table_statement = sqliteparser.parse(sql)[0]
+                    print(create_table_statement)
+                except sqliteparser.SQLiteParserError:
+                    # If sqliteparser can't parse the table schema, just directly print
+                    # what SQLite returned to us.
+                    print(sql)
+        else:
+            if as_python:
+                raise NotImplementedError
+            else:
+                rows = db.select(
+                    "sqlite_master",
+                    where="type = 'table' AND name NOT LIKE 'sqlite_%'",
+                    order_by="name",
+                )
+
+            if rows:
+                print("\n".join(row["name"] for row in rows))
+
+
+@cli.command(name="search")
+@click.argument("db_path")
+@click.argument("table")
+@click.argument("query")
+@click.option("-w", "--where", default="")
+@click.option("--columns", multiple=True, default=[], help=HELP_COLUMNS)
+@click.option("--hide", multiple=True, default=[], help=HELP_HIDE)
+@click.option("-p", "--page", default=1, help=HELP_PAGE)
+@click.option("--limit", default=None, help=HELP_LIMIT)
+@click.option("--offset", default=None, help=HELP_OFFSET)
+@click.option("--order-by", multiple=True, default=[], help=HELP_ORDER_BY)
+@click.option("--desc", is_flag=True, default=False, help=HELP_DESC)
+@click.option(
+    "--plain-foreign-keys", is_flag=True, default=False, help=HELP_PLAIN_FOREIGN_KEYS
+)
+def main_search(
+    db_path,
+    table,
+    query,
+    *,
+    where,
+    columns,
+    hide,
+    page,
+    limit,
+    offset,
+    order_by,
+    desc,
+    plain_foreign_keys,
+):
+    """
+    Alias for 'list <table> -s <query>'.
+    """
+    _select(
+        db_path,
+        table,
+        where=where,
+        search=query,
+        columns=columns,
+        hide=hide,
+        page=page,
+        limit=limit,
+        offset=offset,
+        order_by=order_by,
+        desc=desc,
+        plain_foreign_keys=plain_foreign_keys,
+    )
+
+
 @cli.command(name="sql")
 @click.argument("db_path")
 @click.argument("query")
@@ -697,11 +910,22 @@ def main_sql(db_path, query, *, columns, hide, page, write):
     """
     readonly = not write
     with Database(db_path, readonly=readonly) as db:
-        rows = db.sql(query)
+        try:
+            rows = db.sql(query)
+        except sqlite3.OperationalError as e:
+            if e.args and e.args[0] == "attempt to write a readonly database":
+                report_error_and_exit(
+                    "cannot write to read-only database. Re-run with the --write flag."
+                )
+
+            raise e
+
         if rows:
             prettyprint_rows(rows, columns=columns, hide=hide, page=page)
-        else:
-            print("No rows found.")
+
+        # Don't print a message if there are no rows, because some SQL queries (e.g.,
+        # UPDATE) are not expected to return any rows, and printing a message can be
+        # confusing.
 
 
 @cli.command(name="update")
@@ -712,7 +936,7 @@ def main_sql(db_path, query, *, columns, hide, page, write):
 @click.option(
     "--auto-timestamp/--no-auto-timestamp",
     default=True,
-    help="Automatically populate `last_updated_at` column with current time.",
+    help=HELP_AUTO_TIMESTAMP_UPDATE,
 )
 def main_update(db_path, table, pk, payload, *, auto_timestamp):
     """
@@ -727,11 +951,11 @@ def main_update(db_path, table, pk, payload, *, auto_timestamp):
 
     payload_as_map = {}
     for key_value in payload:
-        key, value = key_value.split("=")
+        key, value = key_value.split("=", maxsplit=1)
         payload_as_map[key] = value
 
     if auto_timestamp:
-        auto_timestamp_columns = ["created_at", "last_updated_at"]
+        auto_timestamp_columns = ["last_updated_at"]
     else:
         auto_timestamp_columns = []
 
@@ -833,7 +1057,16 @@ def get_schema_from_path(schema_path):
 
     schema_module = importlib.util.module_from_spec(spec)
     spec.loader.exec_module(schema_module)
-    return schema_module.SCHEMA
+
+    try:
+        schema = schema_module.SCHEMA
+    except AttributeError:
+        report_error_and_exit(f"did not find SCHEMA variable in {schema_path}")
+
+    if not isinstance(schema, Schema):
+        report_error_and_exit("SCHEMA must be an isqlite.Schema object")
+
+    return schema
 
 
 def group_diff_by_table(diff):
