@@ -8,6 +8,12 @@ import sqliteparser
 from sqliteparser import quote
 
 from . import migrations
+from .exceptions import (
+    ColumnDoesNotExistError,
+    ISqliteApiError,
+    ISqliteError,
+    TableDoesNotExistError,
+)
 from .schema import Diff, Schema, diff_schemas, diff_tables, rename_column
 
 CURRENT_TIMESTAMP_SQL = "STRFTIME('%Y-%m-%d %H:%M:%f', 'now')"
@@ -58,6 +64,7 @@ class Database:
         readonly: Optional[bool] = None,
         uri: bool = False,
         cached_statements: int = 100,
+        enforce_foreign_keys: bool = True,
         insert_auto_timestamp_columns: List[str] = [],
         update_auto_timestamp_columns: List[str] = [],
     ) -> None:
@@ -91,6 +98,8 @@ class Database:
         :param uri: If true, the ``path`` argument is interpreted as a URI rather than a
             file path.
         :param cached_statements: Passed on to ``sqlite3.connect``.
+        :param enforce_foreign_keys: If true, foreign-key constraint enforcement will be
+            turned out with `PRAGMA foreign_keys = 1`.
         :param insert_auto_timestamp_columns: A default value for
             ``auto_timestamp_columns`` in ``insert`` and ``insert_many``. Usually set to
             ``["created_at", "last_updated_at"]`` in conjunction with a schema defined
@@ -139,9 +148,11 @@ class Database:
         self.connection.row_factory = ordered_dict_row_factory
         self.cursor = self.connection.cursor()
 
-        # This must be executed outside a transaction, according to the official
-        # SQLite docs: https://sqlite.org/pragma.html#pragma_foreign_keys
-        self.sql("PRAGMA foreign_keys = 1")
+        if enforce_foreign_keys:
+            # This must be executed outside a transaction, according to the official
+            # SQLite docs: https://sqlite.org/pragma.html#pragma_foreign_keys
+            self.sql("PRAGMA foreign_keys = 1")
+
         if transaction:
             self.sql("BEGIN")
 
@@ -151,6 +162,7 @@ class Database:
         self,
         table: str,
         *,
+        columns: List[str] = [],
         where: str = "",
         values: Dict[str, Any] = {},
         limit: Optional[int] = None,
@@ -165,6 +177,8 @@ class Database:
         :param table: The database table to query. WARNING: This value is directly
             interpolated into the SQL statement. Do not pass untrusted input, to avoid
             SQL injection attacks.
+        :param columns: The columns of the table to return. By default, all columns are
+            returned.
         :param where: A 'where' clause to restrict the query, as a string. The initial
             ``WHERE`` keyword should be omitted. To interpolate Python values, put,
             e.g., ``:placeholder`` in the SQL and then pass ``{"placeholder": x}`` as
@@ -215,15 +229,22 @@ class Database:
         where_clause = f"WHERE {where}" if where else ""
 
         if get_related:
-            columns, joins = self._get_related_columns_and_joins(table, get_related)
+            selection, joins = self._get_related_columns_and_joins(
+                table, columns, get_related
+            )
             rows = self.sql(
-                f"SELECT {columns} FROM {quote(table)} {joins} {where_clause}"
+                f"SELECT {selection} FROM {quote(table)} {joins} {where_clause}"
                 + f" {order_clause} {limit_clause}",
                 values,
             )
         else:
+            if columns:
+                selection = ", ".join(map(quote, columns))
+            else:
+                selection = "*"
+
             rows = self.sql(
-                f"SELECT * FROM {quote(table)} {where_clause} {order_clause}"
+                f"SELECT {selection} FROM {quote(table)} {where_clause} {order_clause}"
                 + f" {limit_clause}",
                 values,
             )
@@ -234,6 +255,7 @@ class Database:
         self,
         table: str,
         *,
+        columns: List[str] = [],
         where: str = "",
         values: Dict[str, Any] = {},
         order_by: Optional[str] = None,
@@ -251,6 +273,7 @@ class Database:
         :param table: The database table to query. WARNING: This value is directly
             interpolated into the SQL statement. Do not pass untrusted input, to avoid
             SQL injection attacks.
+        :param columns: Same as for ``Database.select``.
         :param where: Same as for ``Database.select``.
         :param values: Same as for ``Database.select``.
         :param order_by: Same as for ``Database.select``.
@@ -259,6 +282,7 @@ class Database:
         """
         rows = self.select(
             table,
+            columns=columns,
             where=where,
             values=values,
             order_by=order_by,
@@ -269,7 +293,12 @@ class Database:
         return rows[0] if rows else None
 
     def get_by_pk(
-        self, table: str, pk: int, *, get_related: Union[List[str], bool] = []
+        self,
+        table: str,
+        pk: int,
+        *,
+        columns: List[str] = [],
+        get_related: Union[List[str], bool] = [],
     ) -> Optional[Row]:
         """
         Retrieve a single row from the database table by its primary key.
@@ -278,11 +307,13 @@ class Database:
             interpolated into the SQL statement. Do not pass untrusted input, to avoid
             SQL injection attacks.
         :param pk: The primary key of the row to return.
+        :param columns: Passed on to ``Database.get``.
         :param get_related: Passed on to ``Database.get``.
         """
         pk_column = f"{quote(table)}.rowid"
         return self.get(
             table,
+            columns=columns,
             where=f"{pk_column} = :pk",
             values={"pk": pk},
             get_related=get_related,
@@ -296,6 +327,10 @@ class Database:
         Not to be confused with ``insert_and_get``, which unconditionally inserts a row
         and returns it.
 
+        The returned ``collections.OrderedDict`` object will have an additional
+        ``inserted`` attribute that indicates whether or not a new row was inserted into
+        the database.
+
         :param table: The database table to query. WARNING: This value is directly
             interpolated into the SQL statement. Do not pass untrusted input, to avoid
             SQL injection attacks.
@@ -313,12 +348,12 @@ class Database:
         query = " AND ".join(f"{key} = :{key}" for key in data)
         row = self.get(table, where=query, values=data)
         if row is None:
-            pk = self.insert(table, data, **kwargs)
-            row = self.get_by_pk(table, pk)
-            assert row is not None
-            return row
+            row = self.insert_and_get(table, data, **kwargs)
+            row.inserted = True  # type: ignore
         else:
-            return row
+            row.inserted = False  # type: ignore
+
+        return row
 
     def count(
         self,
@@ -407,6 +442,7 @@ class Database:
         table: str,
         data: Row,
         *,
+        columns: List[str] = [],
         auto_timestamp_columns: Union[List[str], bool] = True,
         get_related: Union[List[str], bool] = [],
     ) -> Row:
@@ -425,7 +461,7 @@ class Database:
           omitted from ``data``.
         """
         pk = self.insert(table, data, auto_timestamp_columns=auto_timestamp_columns)
-        row = self.get_by_pk(table, pk, get_related=get_related)
+        row = self.get_by_pk(table, pk, columns=columns, get_related=get_related)
         assert row is not None
         return row
 
@@ -487,9 +523,9 @@ class Database:
         where: str = "",
         values: Dict[str, Any] = {},
         auto_timestamp_columns: Union[List[str], bool] = True,
-    ) -> None:
+    ) -> int:
         """
-        Update an existing row.
+        Update existing rows and return the number of rows updated.
 
         :param table: The database table. WARNING: This value is directly interpolated
             into the SQL statement. Do not pass untrusted input, to avoid SQL injection
@@ -535,10 +571,11 @@ class Database:
         if self.debugger:
             self.debugger.execute(sql, values)
         self.cursor.execute(sql, values)
+        return self.cursor.rowcount
 
-    def update_by_pk(self, table: str, pk: int, data: Row, **kwargs) -> None:
+    def update_by_pk(self, table: str, pk: int, data: Row, **kwargs) -> bool:
         """
-        Update a single row.
+        Update a single row and return whether it was updated or not.
 
         :param table: The database table. WARNING: This value is directly interpolated
             into the SQL statement. Do not pass untrusted input, to avoid SQL injection
@@ -548,12 +585,14 @@ class Database:
         :param kwargs: Additional arguments to pass on to ``Database.update``.
         """
         pk_column = f"{quote(table)}.rowid"
-        return self.update(
-            table,
-            data,
-            where=f"{pk_column} = :pk",
-            values={"pk": pk},
-            **kwargs,
+        return bool(
+            self.update(
+                table,
+                data,
+                where=f"{pk_column} = :pk",
+                values={"pk": pk},
+                **kwargs,
+            )
         )
 
     def delete(self, table: str, *, where: str, values: Dict[str, Any] = {}) -> None:
@@ -778,13 +817,16 @@ class Database:
         )
         self.refresh_schema()
 
-    def diff(self, schema: Schema, *, table="") -> Diff:
+    def diff(self, schema: Schema, *, table="", detect_renaming=True) -> Diff:
         """
         Return a list of differences between the Python schema and the actual database
         schema.
 
         :param schema: The Python schema to compare against the database.
         :param table: The table to diff. If empty, the entire database will be diffed.
+        :param detect_renaming: If true, the differ will attempt to detect renamed
+            columns. Sometimes the differ does not detect renames correctly, so this
+            option is available to disable renaming detection.
         """
         self.refresh_schema()
         if table:
@@ -794,9 +836,9 @@ class Database:
             except KeyError:
                 raise TableDoesNotExistError(table)
 
-            return diff_tables(old_schema, new_schema)
+            return diff_tables(old_schema, new_schema, detect_renaming=detect_renaming)
         else:
-            return diff_schemas(self.schema, schema)
+            return diff_schemas(self.schema, schema, detect_renaming=detect_renaming)
 
     def apply_diff(self, diff: Diff) -> None:
         """
@@ -843,7 +885,7 @@ class Database:
 
             self.refresh_schema()
 
-    def migrate(self, schema: Schema) -> None:
+    def migrate(self, schema: Schema, *, detect_renaming=True) -> None:
         """
         Migrate the database to match the Python schema.
 
@@ -853,8 +895,9 @@ class Database:
         The entire operation will occur in a transaction.
 
         :param schema: The Python schema to compare against the database.
+        :param detect_renaming: Passed on to ``Database.diff``.
         """
-        self.apply_diff(self.diff(schema))
+        self.apply_diff(self.diff(schema, detect_renaming=detect_renaming))
 
     def refresh_schema(self) -> None:
         """
@@ -973,8 +1016,12 @@ class Database:
         self.sql("PRAGMA foreign_key_check")
 
     def _get_related_columns_and_joins(
-        self, table: str, get_related: Union[List[str], bool]
+        self,
+        table: str,
+        columns_to_select: List[str],
+        get_related: Union[List[str], bool],
     ) -> Tuple[str, str]:
+        # Normalize `get_related` to a set of column names.
         table_schema = self.schema[table]
         if isinstance(get_related, bool):
             if get_related is True:
@@ -994,6 +1041,9 @@ class Database:
         columns_list = []
         joins_list = []
         for column in table_schema.columns:
+            if columns_to_select and column.name not in columns_to_select:
+                continue
+
             if column.name in get_related_set:
                 # Remove the column from the set so that we can check for any
                 # non-existent columns at the end.
@@ -1154,19 +1204,3 @@ class Debugger:
         print(textwrap.indent(f"Values: {values!r}", "  "))
         print()
         print("=== END SQL DEBUGGER ===")
-
-
-class ISqliteError(Exception):
-    pass
-
-
-class ISqliteApiError(ISqliteError):
-    pass
-
-
-class ColumnDoesNotExistError(ISqliteError):
-    pass
-
-
-class TableDoesNotExistError(ISqliteError):
-    pass
