@@ -7,6 +7,7 @@ import sqliteparser
 from . import migrations
 from .columns import primary_key as primary_key_column
 from .columns import timestamp as timestamp_column
+from .exceptions import ColumnDoesNotExistError, TableDoesNotExistError
 
 # Type alias
 Diff = List[migrations.MigrateOperation]
@@ -39,7 +40,10 @@ class Table:
         return cls(stmt.name, stmt.columns)
 
     def __getitem__(self, key: str) -> sqliteparser.ast.Column:
-        return self._columns[key]
+        try:
+            return self._columns[key]
+        except KeyError:
+            raise ColumnDoesNotExistError(key)
 
     def __contains__(self, key: str) -> bool:
         return key in self._columns
@@ -85,7 +89,10 @@ class Schema:
         self._tables = collections.OrderedDict((table.name, table) for table in tables)
 
     def __getitem__(self, key: str) -> Table:
-        return self._tables[key]
+        try:
+            return self._tables[key]
+        except KeyError:
+            raise TableDoesNotExistError(key)
 
     def __contains__(self, key: str) -> bool:
         return key in self._tables
@@ -105,7 +112,9 @@ class Schema:
         return list(self._tables.keys())
 
 
-def diff_schemas(old_schema: Schema, new_schema: Schema) -> Diff:
+def diff_schemas(
+    old_schema: Schema, new_schema: Schema, *, detect_renaming=True
+) -> Diff:
     tables_in_old_schema = set(old_schema.table_names)
     tables_in_new_schema = set(new_schema.table_names)
 
@@ -129,12 +138,12 @@ def diff_schemas(old_schema: Schema, new_schema: Schema) -> Diff:
     for table_name in tables_to_alter:
         old_table = old_schema[table_name]
         new_table = new_schema[table_name]
-        diff.extend(diff_tables(old_table, new_table))
+        diff.extend(diff_tables(old_table, new_table, detect_renaming=detect_renaming))
 
     return diff
 
 
-def diff_tables(old_table: Table, new_table: Table) -> Diff:
+def diff_tables(old_table: Table, new_table: Table, *, detect_renaming=True) -> Diff:
     # TODO(2021-10-17): Clean up this implementation, similar to `diff_schemas`.
     table_name = new_table.name
     diff: Diff = []
@@ -147,8 +156,16 @@ def diff_tables(old_table: Table, new_table: Table) -> Diff:
     for new_index, column in enumerate(new_table.columns):
         old_index = old_columns_to_index_map.get(column.name)
         if old_index is None:
-            if new_index < len(old_table.columns) and is_renamed_column(
-                column, old_table.columns[new_index]
+            if (
+                # Renaming detection must be on...
+                detect_renaming
+                # ...the index must be valid in the old table...
+                and new_index < len(old_table.columns)
+                # ...the column at the index in the old table must be gone...
+                and old_table.columns[new_index].name not in new_table
+                # ...and the old and new columns at that index must otherwise
+                # correspond.
+                and is_renamed_column(column, old_table.columns[new_index])
             ):
                 old_column_name = old_table.columns[new_index].name
                 renamed_columns.add(old_column_name)
